#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "portable_fgets.h"

/*
              F L A S H  1.8
 PRODUCED BY IEEE WORKING GROUP ON ESTIMATING
 LIGHTNING PERFORMANCE OF TRANSMISSION LINES
 ------------------------------------------------
 BASED ON EPRI METHOD BY JOHN ANDERSON
 ORIGINAL FORTRAN METHOD BY NATASHA ROUKOS
 FIRST BASIC VERSION WRITTEN BY JIM WHITEHEAD 9/85
 MODIFIED FOR IBM-PC BY BILL CHISHOLM 1/86 (416)231-4111 X6982
 VERSION 1.4 BY JIM WHITEHEAD  9/87
 VERSION 1.5 BY ROGER CLAYTON AND BILL CHISHOLM 1/89
 VERSION 1.6 BY Ding Yuen and Bill Chisholm 1/90
 VERSION 1.7 FROM WG DEC 1996
 VERSION 1.8 February 2001
 VERSION 1.81 July 2001
    - fixed negative shield angles
	- fixed the Sargent cone surge impedance
    - warning if shielding can't be solved
    - compiled under both Borland C++ Builder 5 and Visual C++ 6
	- note the xls front end doesn't always work with the BCB5 build
*/

enum eUnits {
	eEnglish = 1,
	eMetric
};

enum eACDC {
	eAC,
	eDC
};

enum eTower {
	eConical = 1,
	eHFrame,
	eCylindrical,
	eWaist
};

enum eTower eTowerFirst = eConical;
enum eTower eTowerLast = eWaist;

#define PI     3.14159       /* numerically the same as in v1.7 */
#define FT_PER_M     3.28
#define MM_PER_INCH  25.4
#define INCH_PER_M   39.37
#define KM_PER_MILE  1.609

#define MAX_PHASES  12
#define MAX_SHIELDW  2
#define MAX_EXPOSED  4
#define MAX_ANGLES  36   /* must divide 360 into integer increments */
#define MAX_FR      50
#define MAX_PLUS1   (MAX_PHASES+1)
#define MAX_SECOND  20   /* only using up to 10 in v1.7 ? */

char title [4096];

enum eACDC ACDC [MAX_PHASES];    /* AC or DC? Note 180-deg phase for DC = negative */
double CF   [MAX_PHASES];    /* Coupling Factor */
double CX   [MAX_PHASES];    /* Conductor Horizontal Distance */
double CY   [MAX_PHASES];    /* Conductor Vertical Distance */
double CZ   [MAX_PHASES][2]; /* Mutual Impedance Among Conductors */
double ET   [MAX_PHASES];    /* Effective Tower Flashes For Each Conductor */
double F2   [MAX_PHASES];    /* Flashover Voltage At 2 Microseconds */
double F6   [MAX_PHASES];    /* Flashover Voltage At 6 Microseconds */
double I2   [MAX_PHASES];    /* Critical Stroke Current At 2 Microseconds */
double I6   [MAX_PHASES];    /* Critical Stroke Current At 6 Microseconds */
double IA   [MAX_PHASES];    /* Variable Used In Computing Average Phase Current */
double IC   [MAX_PHASES];    /* Average Value Of Phase Current */
double IR   [MAX_PHASES];    /* Minimum Phase Currents */
double IS   [MAX_PHASES];    /* Lowest Critical Current */
double IU   [MAX_PHASES];    /* Roundoff Value Of IS(I) */
double LV   [MAX_PHASES];    /* Conductor Voltage */
int    PA   [MAX_PHASES];    /* Conductor Phase Angle */
double PH   [MAX_PHASES];    /* Phase Angle Used In Computing Critical Current */
double PR   [MAX_PHASES];    /* Percent Of Time Each Phase Dominates */
double PX   [MAX_PHASES];    /* Probability Of Flashover By Each Conductor */
double SA   [MAX_PHASES];    /* Summing Array For Angles */
double SI   [MAX_PHASES];    /* Strike Distance */
double SU   [MAX_PHASES];    /* Summing Array For Average Current */
double TA   [MAX_PHASES];    /* Travel Time For Each Crossarm */
double TH   [MAX_PHASES];    /* Theta */
double VC   [MAX_PHASES];    /* Voltage Across Crossarm */
double VI   [MAX_PHASES];    /* Insulator Voltage */
double VJ   [MAX_PHASES];    /* Roundoff of VI(I) */
double VP   [MAX_PHASES];    /* Roundoff of VS(I) */
double VR   [MAX_PHASES];    /* Voltage Corresponding To Lowest Critical Current */
double VS   [MAX_PHASES];    /* Insulator Voltage At 6 Microseconds */

double GX   [MAX_SHIELDW];   /* Shieldwire Horizontal Distance */
double GY   [MAX_SHIELDW];   /* Shieldwire Vertical Distance */

double AL   [MAX_EXPOSED];   /* Effective Shield Angle */
double AS1  [MAX_EXPOSED];   /* Actual Shield Angle */
int    MX   [MAX_EXPOSED];   /* Index Of Exposed Conductor */
int    MY   [MAX_EXPOSED];   /* Index Of Shieldwire Protecting Exposed Conductor */

int A  [MAX_PLUS1];             /* Array Used In Computing Critical Current */
int KE [MAX_PLUS1][MAX_SECOND]; /* Matrix Of Phase Angles At End Of Domination */
int KS [MAX_PLUS1][MAX_SECOND]; /* Matrix Of Phase Angles At Start Of Domination */

double SV   [MAX_ANGLES];;   /* Sine Values For Every 10 Degrees */

double RF   [MAX_FR];       /* Footing Resistance Values */
double PRF  [MAX_FR];       /* Probability of Footing Resistance Values 0..1 */

double A1;   /* Angle Used In Computing Uncovered Width */
double A2;   /* Angle Used In Computing Uncovered Width */
double AF;   /* Number of Shielding Failures */
double AM;   /* Variable Used In Computing Mutual Impedances */
double AO;   /* Variable Used In Computing Uncovered Width */
double AP;   /* Footing Resistance Damping */
int B;       /* A(I) */
int BB;      /* A(I) */
double BE;   /* Beta Factor */
double BM;   /* Variable Used In Computing Mutual Impedances */
double BS;   /* Bundle Spacing */
double BU;   /* Equivalent Bundle Radius */
double BW;   /* Horizontal separation between two shield wires */
int    CC;   /* Test Variable For Critical Current */
double CD;   /* Conductor Diameter, mm */
double CS;   /* Conductor Midspan Sag */
double D2;   /* Dummy Variable Used In Computing Corona Radius */
double D9;   /* Dummy Variable Used In Computing Reflected Voltage */
double EC;   /* Equivalent Bundle Radius */
double EF;   /* Effective Tower Flashes */
double EM;   /* Average Shieldwire Height */
double F;    /* Variable Used In Computing Uncovered Width */
double FR;   /* Structure Footing Resistance */
double FV;   /* Flashover Voltage */
double GC;   /* Surge Impedance Of Shieldwire */
double GD;   /* Shieldwire Diameter, mm */
double GFD;  /* Ground Flash Density */
double GM;   /* Variable Used In Computing Combined Surge Impedance */
double GS;   /* Shieldwire Midspan Sag */
double GZ;   /* Surge Impedance Of Shieldwire */
double IL;   /* Initial Guess Of Minimum Current */
double IN;   /* Minimum Current */
enum eTower IT;   /* Tower Type */
double IX;   /* Maximum Current */
enum eUnits MM;   /* Units Selection */
int    NB;   /* Number Of Subconductors */
int    NC;   /* Number Of Conductors */
int    NE;   /* Number Of Shieldwires */
int    NFR;  /* Number Of Footing Resistance Values */
int    NX;   /* Number Of Exposed Conductors */
double P1;   /* Probability Of Maximum Current */
double PB;   /* Probability Of Minimum Current */
double PS;   /* Tower Damping Factor */
double PV;   /* Square Root Of 2/3 */
double RC;   /* Shieldwire Corona Radius */
double RH;   /* Mean Height Of Conductor */
double RN;   /* Strokes To Line/ 100 km */
int    SB;   /* Integer Of SN */
double SD;   /* Strike Distance */
double SM;   /* Maximum Strike Distance */
double SN;   /* Shielding Failure Rate */
double SP;   /* Span Length */
double SS;   /* Bachflashover Rate for a footing resistance */
double SSSUM;/* Bachflashover Rate over the footing resistance histogram */
double SZ;   /* Surge Impedance Of Conductor */
double T1;   /* Angle Variable Used In Computing Average Phase Currents */
double T1H;  /* Variable Used In Computing H-Frame Travel Time */
double T2;   /* Angle Variable Used In Computing Average Phase Currents */
double T2H;  /* Variable Used In Computing H-Frame Travel Time */
double THD;  /* Thunderday Level */
double TM;   /* Distance Used In Computing Effective Shield Angle */
double TPH;  /* Distance from Middle to Top (Type 4 Tower Only) */
double TOPR; /* Tower Top Raduis (Type 4 Tower Only) */
double TR;   /* Tower Base Radius */
double TS;   /* Span Travel Time */
double TU;   /* Tower Travel Time */
double TV;   /* Tower Top Voltage */
double TW;   /* Tower Height */
double V2;   /* Tower Top Voltage At 2 Microseconds */
double V6;   /* Tower Top Voltage At 6 Microseconds */
double VL;   /* Velocity Of Light */
double VT;   /* Tower Top Voltage */
double VV;   /* Reflected Voltage At 6 Microseconds */
double VW;   /* Voltage Across Footing Resistance */
double WW;   /* Tower Midsection Width */
double X1;   /* Coordinate Of Shieldwire */
double X2;   /* Coordinate Of Shieldwire */
double XI;   /* IC(I) */
double XG;   /* Coordinate Of Shieldwire */
double XS;   /* Uncovered Width */
double YO;   /* Variable Used In Computing Maximum Strike Distance */
double Z1;   /* Tower Surge Impedance */
double Z2;   /* Tower Surge Impedance */
double ZI;   /* Intrinsic Circuit Impedance */
double ZT;   /* Tower Surge Impedance */
double ZW;   /* Tower Wave Impedance */

double roundoff (double val) {  /* two decimal places for display */
	return ((int) (val * 100.0)) / 100.0;
}

double GetGFD (double td) {
	return 0.04 * pow (td, 1.25);
}

double GetTDofTH (double th) {
	return pow ((0.054 * pow (th, 1.1)), 1.0 / 1.25);
}

double GetFOV2us (double l) {
	return 820.0 * l;
}

double GetFOV6us (double l) {
	return 585.0 * l;
}

double GetTwoRa (double h) {
	return 28.0 * pow (h, 0.6);
}

double GetProbI (double ka) {
	return 1.0 / (1.0 + pow (ka/31.0, 2.6));
}

double GetStrike (double i, double h) { /* h is a stub for later work */
	return 10.0 * pow (i, 0.65);
}

double GetCoronaRadius (double kv, double h) {
	double rc = 0.2;
	double drc = 1.0;
	double d2;
	int i;

	while (drc > 0.002) { /* Flash 1.7 did iterations in batches of 6 */
		for (i = 0; i < 6; i++) {
			d2 = kv / (1500.0 * log (2.0 * h / rc));
			drc = fabs (d2 - rc);
			rc = d2;
		}
	}

	return rc;
}

/* d and s are in mm, return value in m */
double GetBundleRadius (int n, double d, double s) {
	if (n == 2) {
		return sqrt (s * d / 2e6);
	} else if (n == 3) {
		return pow (s * s * d / 2e9, 1.0 / 3.0);
	} else if (n == 4) {
		return pow (sqrt (2.0) * s * s * s * d / 2e12, 0.25);
	}

	return d / 2000.0;
}

void GetSargentConeParameters (double *z, double *t) {
	*z = 30.0 * log (2.0 * (TW * TW + TR * TR / 4.0) / (TR * TR / 4.0));
	*t = TW / (VL * 0.85);
}

void GetHFrameParameters (double *z, double *t) {
	double z1, z2, t1, t2, lt;

	z1 = 60.0 * log (5.657 * TW / TR) - 60.0;
	t1 = TW / VL;
	z2 = (WW * 60.0 * log (4.0 * TW / TR) + TW * z1) / (WW + TW);
	t2 = (WW + TW) / VL;
	lt = t1 * z1 * t2 * z2 / (t1 * z1 + t2 * z2);

	*z = z1 * z2 / (z1 + z2);
	*t = lt / *z;
}

void GetHilemanCylinderParameters (double *z, double *t) {
	*z = 60.0 * log (5.657 * TW / WW) - 60.0;
	*t = TW / (VL * 0.85);
}

void GetChisholmWaistParameters (double *z, double *t) {
	double z1, z2;

	z2 = 0.5 * (TOPR * TPH + WW * TW + TR * (TW - TPH)) / (TW * TW);
	z2 = 0.5 * atan (z2);
	z1 = 60.0 * log (1.0 / tan (z2)) - 60.0 * log (sqrt (2.0));
	
	*z =  sqrt (PI / 4.0) * z1;  /* because it is square, not round */
	*t = TW / (VL * 0.85);
}

double ipow (double arg, int exp) {
	double ret = 1.0;
	int i;

	if (arg == 0.0) {
		return 0.0;
	} else if (exp > 0) {
		for (i = exp; i > 0; i--) ret *= arg;
	} else if (exp < 0) {
		for (i = exp; i < 0; i++) ret /= arg;
	}

	return ret;
}

void ErrorMessage (const char *msg, const char *line) {
	printf ("ERROR - %s\n", msg);
	if (line && line[0]) {
		printf ("last input was - %s\n", line);
	}
}

void MissingTokenMessage (const char *line, int num_required) {
	char buf [256];
	sprintf (buf, "ERROR - need %d input values\n", num_required);
	ErrorMessage (buf, line);
}

void Reset () {
	int i, j;
	
	NB = NC = NE = NX = NFR = 0;

	IT = eConical;
	MM = eMetric;

	PV = sqrt (2.0 / 3.0);
	VL = 300.0;

	for (i = 0; i < MAX_PHASES; i++) {
		ACDC[i] = eAC;
		PA[i] = 0;
		IA[i] = IU[i] = CF[i] = CX[i] = CY[i] = CZ[i][0] = CZ[i][1] = 0.0;
		ET[i] = F2[i] = F6[i] = I2[i] = I6[i] = 0.0;
		IC[i] = IR[i] = IS[i] = LV[i] = PH[i] = 0.0;
		PR[i] = PX[i] = SA[i] = SI[i] = SU[i] = TA[i] = 0.0;
		TH[i] = VC[i] = VI[i] = VJ[i] = VP[i] = VR[i] = VS[i] = 0.0;
	}

	for (i = 0; i < MAX_SHIELDW; i++) {
		GX[i] = GY[i] = 0.0;
	}

	for (i = 0; i < MAX_EXPOSED; i++) {
		MX[i] = MY[i] = 0;
		AL[i] = AS1[i] = 0.0;
	}

	for (i = 0; i < MAX_ANGLES; i++) {
		SV[i] = 0.0;
	}

	for (i = 0; i < MAX_FR; i++) {
		RF[i] = 0.0;
		PRF[i] = 0.0;
	}

	for (i = 0; i < MAX_PLUS1; i++) {
		A[i] = 0;
		for (j = 0; j < MAX_SECOND; j++) {
			KE[i][j] = KS[i][j] = 0;
		}
	}

	A1 = A2 = AF = AM = AO = AP = BE = BM = BS = BU = BW = 0.0;
	CD = CS = D2 = D9 = EC = EF = EM = F = FR = FV = GC = GD = GFD = 0.0;
	GM = GS = GZ = IL = IN = IX = P1 = PB = PS = RC = RH = RN = SD = 0.0;
	SM = SN = SP = SS = SZ = T1 = T1H = T2 = T2H = THD = TM = TPH = 0.0;
	TOPR = TR = TS = TU = TV = TW = V2 = V6 = VT = VV = VW = WW = X1 = 0.0;
	SSSUM = X2 = XI = XG = XS = YO = Z1 = Z2 = ZI = ZT = ZW = 0.0;

	B = BB = CC = SB = 0;
}

void SetBaseCase () {
	int i;

	NC = 3;
	NE = 2;
	
	for (i = 0; i < NC; i++) {
		CX[i] = 5.0 * i - 5.0;
		CY[i] = 35.0;
		SI[i] = 3.0;
		LV[i] = 500.0;
		PA[i] = 120 * i;
		ACDC[i] = eAC;
	}

	for (i = 0; i < NE; i++) {
		GX[i] = 6.0 * i - 3.0;
		GY[i] = 40.0;
	}

	NX = 2;
	MX[0] = 0;
	MY[0] = 0;
	MX[1] = 2;
	MY[1] = 1;

	NB = 4;
	BS = 457.0;
	CD = 30.0;
	CS = 0.0;
	
	GD = 14.0;
	GS = 0.0;

	IT = eWaist;
	TW = 40.0;
	TR = 8.0;
	TPH = 20.0;
	WW = 4.0;
	TOPR = 2.0;

	SP = 300.0;
	THD = 40.0;

	MM = eMetric;

	for (i = 0; i < 10; i++) {
		RF[i] = 100.0 - i * 10.0;
	}
}

void ScaleUnits (double length_scale, double spacing_scale, double diameter_scale) {
	int i;
	
	for (i = 0; i < NC; i++) {
		CX[i] *= length_scale;
		CY[i] *= length_scale;
		SI[i] *= length_scale;
	}

	for (i = 0; i < NE; i++) {
		GX[i] *= length_scale;
		GY[i] *= length_scale;
	}

	BS *= spacing_scale;
	CD *= diameter_scale;
	GD *= diameter_scale;

	TW *= length_scale;
	TR *= length_scale;
	WW *= length_scale;
	TPH *= length_scale;
	SP *= length_scale;
	CS *= length_scale;
	GS *= length_scale;
}

void EnglishToMetric () {
	ScaleUnits (1.0 / FT_PER_M, MM_PER_INCH, MM_PER_INCH);
	GFD /= (KM_PER_MILE * KM_PER_MILE);
}

void MetricToEnglish () {
	ScaleUnits (FT_PER_M, 1.0 / MM_PER_INCH, 1.0 / MM_PER_INCH);
}

int IsComment (char *buf) {
	char c1 = buf[0];

	if (c1 == '!' || c1 == '/') return 1;

	return 0;
}

int IsEmpty (char *buf) {
	int imax = strlen (buf);
	int i;
	
	for (i = 0; i < imax; i++) {
		if (isalnum (buf[i])) return 0;
	}

	return 1;
}

int GetNextDataLine (FILE *fp, char *buf, int maxlen, int required) {
	buf[0] = '\0';

	while (zap_fgets (buf, maxlen, fp)) {
		if (IsComment (buf)) continue;
		if (IsEmpty (buf)) continue;
		return 1;
	}

	if (required) {
		ErrorMessage ("premature end of input data", buf);
	}

	buf[0] = '\0';
	return 0; /* no more lines */
}

int ValidateInteger (int val, int min, int max, const char *label, const char *line)
{
	char msg [256];
	if (val >= min && val <= max) return 1;
	
	sprintf (msg, "value '%s' [%d] must be in the range %d to %d", 
		label, val, min, max);
	ErrorMessage (msg, line);

	return 0;
}

int ValidateDouble (double val, double min, double max, const char *label, const char *line)
{
	char msg [256];
	if (val >= min && val <= max) return 1;
	
	sprintf (msg, "value '%s' [%G] must be in the range %G to %G", 
		label, val, min, max);
	ErrorMessage (msg, line);

	return 0;
}

int ReadFromFile (FILE *fp) {
	char buf [4096];
	char tok [4096];
	int i;
	int nFRTokens;
	double sumPRF;

	/* read the case title */
	if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
	strncpy (title, buf, sizeof title);

	/* read the units and conductor counts */
	if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
	if (3 != sscanf (buf, "%d %d %d", &MM, &NC, &NE)) {
		MissingTokenMessage (buf, 3);
		return 0;
	}
	if (!ValidateInteger (MM, eEnglish, eMetric, "MM", buf)) return 0;
	if (!ValidateInteger (NC, 1, MAX_PHASES, "NC", buf)) return 0;
	if (!ValidateInteger (NE, 1, MAX_SHIELDW, "NE", buf)) return 0;

	/* read the conductor geometry and voltage levels */
	for (i = 0; i < NC; i++) {
		if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
		if (6 != sscanf (buf, "%lf %lf %lf %lf %d %s", 
			&CX[i], &CY[i], &SI[i], &LV[i], &PA[i], tok)) {
			MissingTokenMessage (buf, 6);
			return 0;
		}
		if (tok[0] == 'D' || tok[0] == 'd') {
			ACDC[i] = eDC;
		} else {
			ACDC[i] = eAC;
		}
		if (!ValidateDouble (CX[i], -200.0, 200.0, "CX", buf)) return 0;
		if (!ValidateDouble (CY[i], 0.1, 200.0, "CY", buf)) return 0;
		if (!ValidateDouble (SI[i], 0.1, 200.0, "SI", buf)) return 0;
		if (!ValidateDouble (LV[i], 0.1, 2000.0, "LV", buf)) return 0;
		if (!ValidateInteger (PA[i], -360, 360, "PA", buf)) return 0;
	}

	/* read the shield wire positions */
	for (i = 0; i < NE; i++) {
		if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
		if (2 != sscanf (buf, "%lf %lf", &GX[i], &GY[i])) {
			MissingTokenMessage (buf, 2);
			return 0;
		}
		if (!ValidateDouble (GX[i], -200.0, 200.0, "GX", buf)) return 0;
		if (!ValidateDouble (GY[i], 0.1, 200.0, "GY", buf)) return 0;
	}

	/* read the number of exposed conductors */
	if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
	if (1 != sscanf (buf, "%d", &NX)) {
		MissingTokenMessage (buf, 1);
		return 0;
	}
	if (!ValidateInteger (NX, 1, MAX_EXPOSED, "NX", buf)) return 0;

	/* read the exposed conductor - shield wire pairs */
	for (i = 0; i < NX; i++) {
		if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
		if (2 != sscanf (buf, "%d %d", &MX[i], &MY[i])) {
			MissingTokenMessage (buf, 2);
			return 0;
		}
		if (!ValidateInteger (MX[i], 1, NC, "MX", buf)) return 0;
		if (!ValidateInteger (MY[i], 1, NE, "MY", buf)) return 0;
		--MX[i]; /* C arrays are zero-based */
		--MY[i];
	}

	/* read the bundling, conductor diameter, and conductor sags */
	if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
	if (4 != sscanf (buf, "%d %lf %lf %lf", &NB, &BS, &CD, &CS)) {
		MissingTokenMessage (buf, 4);
		return 0;
	}
	if (!ValidateInteger (NB, 1, 4, "NB", buf)) return 0;
	if (!ValidateDouble (BS, 0.0, 1000.0, "BS", buf)) return 0;
	if (!ValidateDouble (CD, 0.0, 100.0, "CD", buf)) return 0;
	if (!ValidateDouble (CS, 0.0, 100.0, "CS", buf)) return 0;
	
	/* read the shield wire diameter, sag, span, and thunder-day level */
	if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
	if (4 != sscanf (buf, "%lf %lf %lf %lf", &GD, &GS, &SP, &THD)) {
		MissingTokenMessage (buf, 4);
		return 0;
	}
	if (!ValidateDouble (GD, 0.0, 100.0, "GD", buf)) return 0;
	if (!ValidateDouble (GS, 0.0, 100.0, "GS", buf)) return 0;
	if (!ValidateDouble (SP, 0.0, 2000.0, "SP", buf)) return 0;
	if (THD > 0.0) {
		if (!ValidateDouble (THD, 0.0, 100.0, "THD", buf)) return 0;
		GFD = GetGFD (THD);
	} else {
		GFD = -THD;
		if (!ValidateDouble (GFD, 0.0, 100.0, "GFD", buf)) return 0;
	}

	/* estimated sags */
	if (CS <= 0.0) {
		if (MM == eEnglish) {
			CS = 0.0014 * SP * SP / 20.0;
		} else {
			CS = 2.1336e-5 * SP * SP;
		}
	}
	if (GS <= 0.0) {
		if (MM == eEnglish) {
			GS = 0.0009 * SP * SP / 20.0;
		} else {
			GS = 1.3716e-5 * SP * SP;
		}
	}
	
	/* read the tower data */
	if (!GetNextDataLine (fp, buf, sizeof buf, 1)) return 0;
	if (6 != sscanf (buf, "%d %lf %lf %lf %lf %lf", 
		&IT, &TW, &TR, &WW, &TPH, &TOPR)) {
		MissingTokenMessage (buf, 6);
		return 0;
	}
	if (!ValidateInteger (IT, eTowerFirst, eTowerLast, "IT", buf)) return 0;
	if (!ValidateDouble (TW, 0.0, 300.0, "TW", buf)) return 0;
	if (!ValidateDouble (TR, 0.0, 100.0, "TR", buf)) return 0;
	if (!ValidateDouble (WW, 0.0, 100.0, "WW", buf)) return 0;
	if (!ValidateDouble (TPH, 0.0, 100.0, "TPH", buf)) return 0;
	if (!ValidateDouble (TOPR, 0.0, 100.0, "TOPR", buf)) return 0;

	/* tower pre-processing */
	
	/* read the footing resistance 1-cdf distribution, exactly 10 values */
	NFR = 0;
	nFRTokens = 0; /* not yet determined */
	sumPRF = 0.0;

	while (GetNextDataLine (fp, buf, sizeof buf, 0)) {
		++NFR;
		if (NFR > MAX_FR) {
			sprintf (buf, "Too many footing resistance values, limit is %d\n", MAX_FR);
			ErrorMessage (buf, NULL);
			return 0;
		}
		i = NFR - 1;

		/* decide if reading cdf (1 token) or histogram (2 tokens) */
		if (nFRTokens <= 0) {
			nFRTokens = 1;
			if (2 == sscanf (buf, "%lf %lf", &PRF[i], &RF[i])) {
				nFRTokens = 2;
			}
		}
		
		if (nFRTokens == 1) {
			if (1 != sscanf (buf, "%lf", &RF[i])) {
				MissingTokenMessage (buf, 1);
				return 0;
			}
			PRF[i] = 10.0;
		} else if (nFRTokens == 2) {
			if (2 != sscanf (buf, "%lf %lf", &PRF[i], &RF[i])) {
				MissingTokenMessage (buf, 2);
				return 0;
			}
		}

		if (!ValidateDouble (PRF[i], 1.0e-3, 100.0e6, "PRF", buf)) return 0;
		if (!ValidateDouble (RF[i], 1.0e-3, 100.0e6, "RF", buf)) return 0;

		sumPRF += PRF[i];
	}

	if (nFRTokens == 1 && NFR != 10) {
		ErrorMessage ("Must have 10 footing resistance values for cumulative distribution.\n", NULL);
		return 0;
	} else if (NFR < 1) {
		ErrorMessage ("Must have at least one footing resistance value for histogram.\n", NULL);
		return 0;
	}

	/* normalize the footing resistance histogram */
	for (i = 0; i < NFR; i++) {
		PRF[i] /= sumPRF;
	}

	return 1;
}

/* for user instructions - this requires the X=0 reference to be tower centerline */
int DoUnshieldedCase (int bWriteOutput) {
	int ileft = 0;
	int iright = 0;
	int i;
	double hleft = 0.0;
	double hright = 0.0;
	double havg;

	/* find the line width */
	for (i = 0; i < NC; i++) {
		if (CX[i] > 0.0 && CY[i] > hright) {
			hright = CY[i];
			iright = i;
		}
		if (CX[i] <= 0.0 && CY[i] > hleft) {
			hleft = CY[i];
			ileft = i;
		}
	}

	/* flashes to unshielded line */
	havg = 0.5 * (hleft + hright);
	BW = fabs (CX[iright] - CX[ileft]);
	RN = 0.1 * GFD * (BW + GetTwoRa (havg));

	/* corona and bundle adjustments */
	FV = 0.5 * (GetFOV6us (SI[ileft]) + GetFOV6us (SI[iright]));
	RC = GetCoronaRadius (FV, RH);
	EC = GetBundleRadius (NB, CD, BS);
	if (NB > 1) RC += EC;

	/* surge impedance and critical current */
	SZ = 60.0 * sqrt (log (2.0 * RH / EC) * log (2.0 * RH / RC));
	IN = 2.0 * FV / SZ;

	/* flashes to unshielded line that cause a flashover */
	SN = GFD * GetProbI (IN);

	if (bWriteOutput) {
		printf ("*** Start Unshielded Line Analysis ***\n");
		printf ("GFD =              %7.2f\n", GFD);
		printf ("Line Width =       %7.2f\n", BW);
		printf ("Flashes to Line =  %7.2f\n", RN);
		printf ("Corona Radius =    %7.2f\n", NB > 1 ? RC - EC : RC);
		printf ("Bundle Radius =    %7.2f\n", EC);
		printf ("Surge Impedance =  %7.2f\n", SZ);
		printf ("Critical Current = %7.2f\n", IN);
		printf ("*** End Unshielded Line Analysis ***\n");
	}

	return 1;
}

/* return -1 to stop the backflash calculation,
   0 to indicate failure of the shielding calculation,
   1 to indicate success */

int DoShielding (int bWriteOutput) {
	int i, ix, iy;
	double GYavg = 0.0;

	/* 1.4 - average shield wire height */
	for (i = 0; i < NE; i++) GYavg += GY[i];
	GYavg /= NE;
	EM = GYavg - (2.0 / 3.0) * GS;

	/* 1.5 - total flashes to line */
	RN = 0.1 * GFD * (fabs (GX[0] - GX[1]) + GetTwoRa (GYavg));

	if (bWriteOutput) {
		printf ("*** Start Shielding Analysis ***\n");
		printf ("GFD =             %7.2f\n", GFD);
		printf ("Flashes to Line = %7.2f\n", RN);
		printf ("Idx    Beta   Icrit\n");
	}

	/* 1.6 - flashover of most exposed conductor */
	SN = 0.0;
	for (i = 0; i < NX; i++) {
		ix = MX[i];
		iy = MY[i];
		FV = GetFOV6us (SI[ix]);
	
		/* 1.7 - mean phase conductor height */
		RH = CY[ix] - (2.0 / 3.0) * CS;

		/* 1.7a - calculate beta from a phase conductor height, 1/1986 */
		if (i == 0) {
			/* old beta formulation */
			BE = 22.0 / RH;
			if (BE > 0.9) BE = 0.9;
			if (BE < 0.6) BE = 0.6;

			/* new beta formulation */
			if (RH > 40.0) {
				BE = 0.55;
			} else {
				BE = 3.6 + 1.7 * log (43.0 - RH);
				BE *= 0.1;
			}
		}

		/* 1.8 - corona radius */
		RC = GetCoronaRadius (FV, RH);

		/* 1.9 and 1.10 - equivalent bundle radius */
		EC = GetBundleRadius (NB, CD, BS);
		if (NB > 1) RC += EC;

		/* 1.11 and 1.12 - effective self surge impedance of phase conductor */
		SZ = 60.0 * sqrt (log (2.0 * RH / EC) * log (2.0 * RH / RC));
		IN = 2.0 * FV / SZ;

		/* 1.13 - striking distance */
		SD = GetStrike (IN, RH);
		
		/* 1.14 and 1.15 - coordinate of shield wire for good shielding */
		X1 = SD * SD - ipow (BE * SD - RH, 2);
		if (X1 <= 0.0) {
			ErrorMessage ("Hand calculate shielding - Rg and Rc don't intersect for phase conductor.", NULL);
			return 0;  /* v1.7 said 'hand calculation necessary' */
		}

		X2 = SD * SD - ipow (BE * SD - EM, 2);
		if (X2 <= 0.0) {
			ErrorMessage ("Hand calculate shielding - Rg and Rc don't intersect for shield wire.", NULL);
			return 0;  /* v1.7 said 'hand calculation necessary' */
		}

		XG = sqrt (X2) - sqrt (X1);

		/* 1.16 - effective shield angle */
		AL[i] = atan (XG / (EM - RH));
		TM = fabs (CX[ix] - GX[iy]) / (EM - RH);
		AS1[i] = atan (TM);
		if (fabs (GX[iy]) > fabs (CX[ix])) AS1[i] *= -1.0;
		
		if ((AS1[i] - AL[i] <= 0.0) || AS1[i] <= 0.0) {
		} else {
			/* 1.17 - compute the uncovered width */
			F = sqrt (ipow (CX[ix] - GX[iy], 2) + ipow (RH - EM, 2));
			AO = F / (2.0 * SD);
			A1 = fabs (AS1[i]) - (PI / 2.0 - atan (AO / sqrt(1.0 - AO * AO)));
			A2 = (BE * SD - RH) / SD;
			if (BE * SD < RH) {
				XS = SD + SD * sin (A1);
			} else {
				XS = SD * cos (atan (A2 / sqrt (1.0 - A2 * A2))) + SD * sin (A1);
			}

			/* 1.18 - maximum strike distance that can occur */
			YO = (EM + RH) / 2.0;
			SM = fabs (1.0 / (TM / sqrt (1.0 + TM * TM) - BE) * YO);

			/* 1.19 and 1.20 - compute Imaximum */
			IX = 0.041 * pow (SM, 1.538);

			/* 1.21 through 1.24 - total line shielding failures */
			PB = GetProbI (IN);
			P1 = GetProbI (IX);
			AF = (GFD / 10.0) * XS * (PB - P1) / 2.0;
			SN += AF;
		}
		AL[i] *= 180.0 / PI;
		AL[i] = roundoff (AL[i]);
		AS1[i] *= 180.0 / PI;
		AS1[i] = roundoff (AS1[i]);
		
		if (bWriteOutput) {
			printf ("%3d %7.2f %7.2f\n", ix+1, BE, IN);
		}
	}

	if (bWriteOutput) {
		printf ("*** End Shielding Analysis   ***\n\n");
	}

	return 1;
}

/* requires NE > 0 */
int DoBackflash (int bWriteOutput) {
	/* double betaS;  not used */
  double ang, d_ang;
	int i, j, jmax, iTFR, LA, k, d_k, NO;
	
	/* 2.10 - span travel time */
	TS = SP / (VL * 0.9);

	/* 2.1 and 2.2 - flashover voltages at 2 and 6 us */
	for (i = 0; i < NC; i++) {
		if (TS >= 1.0) {
			F2[i] = SI[i] * (400.0 + 710.0 * pow (2.0 * TS, -0.75));
		} else {
			F2[i] = GetFOV2us (SI[i]);
		}
		F6[i] = GetFOV6us (SI[i]);
	}

	/* 2.3 - tower-top voltage */
	TV = 1.8 * F2[0]; /* TODO - always the first phase conductor? */

	/* 2.4 - shield wire corona adjustment */
	RC = GetCoronaRadius (TV, GY[0]); /* TODO - always the first shield wire? */
		
	/* 2.5 - self surge impedance of each shield wire */
	GZ = 60.0 * sqrt (log (2.0 * GY[0] / (GD / 2000.0)) * log (2.0 * GY[0] / RC));

	/* 2.6 - combined surge impedance */
	if (NE == 1) {
		GC = GZ;
	} else { /* TODO - hard wired for 2 shield wires */
		BM = ipow (GX[0] - GX[1], 2);
		AM = BM + ipow (2.0 * GY[0], 2);
		GM = 60.0 * log (sqrt (AM / BM));
		GC = (GZ + GM) / 2.0;
	}

	/* 2.7 - mutual impedances between conductors and shield wires */
	for (i = 0; i < NE; i++) {
		for (j = 0; j < NC; j++) {
			AM = BM = ipow (CX[j] - GX[i], 2);
			AM += ipow (CY[j] + GY[i], 2);
			BM += ipow (CY[j] - GY[i], 2);
			CZ[j][i] = 60.0 * log (sqrt (AM/BM));
		}
	}

	/* 2.7b - coupling factors */
	for (i = 0; i < NC; i++) {
		if (NE == 1) {
			CF[i] = CZ[i][0] / GC;
		} else {  /* TODO - hard wired for 2 */
			CF[i] = (CZ[i][0] + CZ[i][1]) / 2.0 / GC;
		}
	}

	/* 2.8 and 2.9 - tower surge response */
	if (IT == eConical) {
		GetSargentConeParameters (&ZT, &TU);
	} else if (IT == eHFrame) {
		GetHFrameParameters (&ZT, &TU);
	} else if (IT == eCylindrical) {
		GetHilemanCylinderParameters (&ZT, &TU);
	} else { /*	eWaist */
		GetChisholmWaistParameters (&ZT, &TU);
	}

	/* 2.11 - travel time to each crossarm */
	for (i = 0; i < NC; i++) {
		TA[i] = (TW - CY[i]) / (VL * 0.85);  /* TODO - parameterize tower wave velocity */
	}

	/* 2.12 and 2.13 - intrinsic circuit impedance */
	ZI = (GC * ZT) / (GC + 2.0 * ZT);

	if (bWriteOutput) {
		printf ("*** Start Backflash Analysis ***\n");
		printf ("  TS = %7.3f\n", TS);
		printf ("  ZT = %7.2f\n", ZT);
		printf ("  TU = %7.4f\n", TU);
		printf ("  TV = %7.2f\n", TV);
		printf ("  RC = %7.4f\n", RC);
		printf ("  EC = %7.4f\n", EC);
		printf ("  GZ = %7.2f\n", GZ);
		printf ("  GC = %7.2f\n", GC);
		printf ("  ZI = %7.2f\n", ZI);
		printf ("  Idx      CF      F2      F6      TA\n");
		for (i = 0; i < NC; i++) {
			printf ("  %3d %7.4f %7.1f %7.1f %7.4f\n", i+1, CF[i], F2[i], F6[i], TA[i]);
		}
	}

	/* start the main footing resistance loop */
	for (iTFR = 0; iTFR < NFR; iTFR++) {
		FR = RF[iTFR];
		
		/* 2.14 - tower wave impedance */
		ZW = (2.0 * ZT * GC * GC) / (ipow (GC + 2.0 * ZT, 2)) * (ZT - FR) / (ZT + FR);

		/* 2.15 - tower damping factor */
		PS = (2.0 * ZT - GC) / (2.0 * ZT + GC) * (ZT - FR) / (ZT + FR);
    /* betaS is not used
		betaS = (2.0 * ZT - GC) / (2.0 * ZT + GC);
    */

		/* 2.16 - footing resistance damping factor */
		AP = (2.0 * FR) / (ZT + FR);

		/* 2.17 through 2.19 - tower top voltage */
		V2 = ZI - ZW * (1.0 - TU / (1.0 - PS)) / (1.0 - PS);
		if (TS < 1.0) {
			VT = -4.0 * 0.85 * V2 * V2 / GC * (1.0 - 2.0 * V2 / GC) * (1.0 - TS);
		} else {
			VT = 0.0;
		}
		V2 += VT;

		/* 2.20 - voltage across footing resistance at 2 us */
		VW = AP * ZI * (1.0 - (PS * TU) / (1.0 - PS)) / (1.0 - PS);

		/* 2.21 - reduce the voltage across footing resistance by reflection */
		VW += VW * VT / V2;

		/* 2.22 - crossarm voltages at 2 us */
		for (i = 0; i < NC; i++) {
			VC[i] = VW + ((TU - TA[i]) / TU) * (V2 - VW);
		}

		/* 2.23 - insulator voltages at 2 us */
		for (i = 0; i < NC; i++) {
			VI[i] = VC[i] - CF[i] * V2;
		}

		/* 2.24 - tower top voltage at 6 us */
		V6 = GC * FR / (GC + 2.0 * FR);

		/* 2.25 - reflected voltage at 6 us */
		D9 = FR / (GC + 2.0 * FR);
		VV = -4.0 * 0.85 * GC * D9 * D9 * (1.0 - 2.0 * D9);

		/* 2.26 - insulator voltages at 6 us */
		for (i = 0; i < NC; i++) {
			VS[i] = (V6 + VV) * (1.0 - CF[i]);
		}

		/* 2.27 - critical stroke current at 2 us */
		for (i = 0; i < NC; i++) {
			I2[i] = F2[i] / VI[i];
		}

		/* 2.28 - critical stroke current at 6 us */
		for (i = 0; i < NC; i++) {
			I6[i] = F6[i] / VS[i];

			/* 2.29 and 2.30 - select lowest critical current */
			if (I2[i] < I6[i]) {
				IS[i] = I2[i];
				VR[i] = F2[i];
			} else {
				IS[i] = I6[i];
				VR[i] = F6[i];
			}
			IU[i] = roundoff (IS[i]);
			VJ[i] = roundoff (VI[i]);
			VP[i] = roundoff (VS[i]);
		}

		/* 2.31 - find equation for critical current */
		d_ang = 360.0 / MAX_ANGLES;
		ang = 0.5 * d_ang;
		for (i = 0; i < MAX_ANGLES; i++) {
			SV[i] = sin (ang * PI / 180.0);
			ang += d_ang;
		}
		for (i = 0; i < MAX_PLUS1; i++) {
			A[i] = 0; /* starting index into a zero-based C array */
			PR[i] = 0.0;
			for (j = 0; j < MAX_SECOND; j++) {
				KS[i][j] = KE[i][j] = 0;
			}
		}
		LA = NC; /* C arrays are zero-based */
		NO = 0;
		d_k = 360 / MAX_ANGLES;
		for (k = d_k / 2; k < 360; k += d_k) {
			IL = 2000.0; /* initial guess of min current [kA] */
			for (j = 0; j < NC; j++) {
				CC = (k + PA[j]) / d_k;
				while (CC < 0) CC += 36; /* TODO - MAX_ANGLES / 10 or 20 ? */
				while (CC > 35) CC -= 36;
				PH[j] = SV[CC];
				if (ACDC[j] == eAC) {
					IR[j] = IS[j] * (1.0 - PV * PH[j] * LV[j] / VR[j]);
				} else {
					IR[j] = IS[j] * (1.0 - cos (PI * PA[j] / 180.0) * LV[j] / VR[j]);
				}
				if (IR[j] <= IL) {
					IL = IR[j];
					NO = j;
				}
			}
			PR[NO] += 1;
			B = A[LA];
			if (LA != NO) {
				KE[LA][B] = k - d_k / 2;
				A[NO] += 1;
				BB = A[NO];
				KS[NO][BB] = k - d_k / 2;
				LA = NO;
			}
		}
		KE[LA][BB] = k - d_k / 2;

		/* 2.32 - percent of time each phase dominates */
		for (i = 0; i < NC; i++) {
			PR[i] *= 100.0 / 36;  /* TODO - MAX_ANGLES / 10 or 20 ? */
		}

		/* 2.33 - average value of current for each phase */
		for (i = 0; i < NC; i++) {
			SU[i] = SA[i] = 0.0;
			jmax = A[i];
			if (jmax < 1) {
				IC[i] = 0.0;
				continue;
			}
			for (j = 1; j <= jmax; j++) {
				T2 = (KE[i][j] + PA[i]) * PI / 180.0;
				T1 = (KS[i][j] + PA[i]) * PI / 180.0;
				TH[j] = T2 - T1;
				if (ACDC[i] == eAC) {
					IA[j] = IS[i] * (1.0 + (PV * LV[i] / VR[i]) * (cos(T2) - cos(T1)) / TH[j]);
				} else {
					IA[j] = IS[i] * (1.0 - cos(PA[i] * PI / 180.0) * LV[i] / VR[i]);
				}
				SU[i] += IA[j] * TH[j];
				SA[i] += TH[j];
			}
			IC[i] = SU[i] / SA[i];
		}

		/* 2.34 - probability that stroke current is exceeded */
		for (i = 0; i < NC; i++) {
			if (IC[i] > 0.0) {
				PX[i] = GetProbI (IC[i]);
			} else {
				PX[i] = 0.0;
			}
		}

		/* 2.35 - effective tower flashes */
		EF = 0.6 * RN;

		/* 2.36 through 2.38 - tower flashes per phase per 100 km per year */
		SS = 0.0;
		for (i = 0; i < NC; i++) {
			ET[i] = EF * PR[i] * PX[i] / 100.0;
			SS += ET[i];
		}

		SS = roundoff (SS);

		SSSUM += SS * PRF [iTFR];

		if (bWriteOutput) {
			printf ("*** Section %1d (%7.2f%% of line)\n", 
				iTFR+1, PRF[iTFR] * 100.0);
			printf ("  FR = %7.2f\n", FR);
			printf ("  ZW = %7.2f\n", ZW);
			printf ("  SS = %7.2f\n", SS);
			printf ("  Idx      I2      I6   Icrit    Prob  IcrAvg\n");
			for (i = 0; i < NC; i++) {
				printf ("  %3d %7.2f %7.2f %7.2f %7.1f %7.2f\n", 
					i+1, VJ[i], VP[i], IU[i], PR[i], IC[i]);
			}
		}

	} /* ends footing resistance loop */

	if (bWriteOutput) {
		printf ("*** End Backflash Analysis   ***\n");
	}

	return 1;
}

int DoCalculations (int bWriteOutput) {
	int ret;

	if (NE < 1) return DoUnshieldedCase (bWriteOutput);

	ret = DoShielding (bWriteOutput);

	if (ret >= 0) {
		return DoBackflash (bWriteOutput);
	}

	return 0;
}

int WriteBriefOutput () {
	int i;

	printf ("%s\n", title);
	printf ("  Flashovers per 100 km per year:\n");
	printf ("    BFR  = %7.2f\n", SSSUM);
	printf ("    SFFR = %7.2f\n", SN);
	printf ("    TFR  = %7.2f\n", SSSUM + SN);

	if (NX > 0) printf ("  Shielding Angles [PhaseIdx, ShieldIdx, Required, Actual]:\n");
	for (i = 0; i < NX; i++) {
		printf ("   %2d %2d %7.2f %7.2f\n", 1+MX[i], 1+MY[i], AL[i], AS1[i]);
	}

	return 1;
}

int WriteSummaryOutput () {
	int i;
	double dscale;

	printf ("\n*** Summary of Results *************************************************\n");
	if (MM == eEnglish) {
		dscale = KM_PER_MILE;
		printf ("Flashovers per 100 miles per year:\n");
	} else {
		dscale = 1.0;
		printf ("Flashovers per 100 km per year:\n");
	}
	printf ("  BFR  = %7.2f\n", SSSUM * dscale);
	printf ("  SFFR = %7.2f\n", SN * dscale);
	printf ("  TFR  = %7.2f\n", (SSSUM + SN) * dscale);

	if (NX > 0) printf ("\nShielding Angles [PhaseIdx, ShieldIdx, Required, Actual]:\n");
	for (i = 0; i < NX; i++) {
		printf (" %2d %2d %7.2f %7.2f\n", 1+MX[i], 1+MY[i], AL[i], AS1[i]);
	}

	return 1;
}

void EchoConductors (const char *label) {
	int i;

	if (label && label[0]) {
		printf ("%s\n", label);
	}

	printf ("Span Length = %7.2f\n", SP);
	printf ("Conductors: Dia = %7.2f, Sag = %7.2f, Nsub = %d, BundSpc = %7.2f\n",
		CD, CS, NB, BS);
	printf ("Idx  Xcoord  Height  InsLen   Voltage  Angle  ACDC\n");
	for (i = 0; i < NC; i++) {
		printf ("%3d %7.2f %7.2f %7.2f   %7.2f %6d    %s\n",
			i+1, CX[i], CY[i], SI[i], LV[i], PA[i], ACDC[i] == eAC ? "AC" : "DC");
	}
	
	printf ("Shield Wires: Dia = %7.2f, Sag = %7.2f\n", 
		GD, GS);
	printf ("Idx  Xcoord  Height\n");
	for (i = 0; i < NE; i++) {
		printf ("%3d %7.2f %7.2f\n", i+1, GX[i], GY[i]);
	}

	if (NX > 0) {
		printf ("Exposed Phase Indices:  ");
		for (i = 0; i < NX; i++) printf ("%3d", MX[i]+1);
		printf ("\n");
		printf ("Shielding Wire Indices: ");
		for (i = 0; i < NX; i++) printf ("%3d", MY[i]+1);
		printf ("\n");
	}

	printf ("Tower Model - ");
	if (IT == eConical) {
		printf ("Cone:\n");
		printf ("  Height        = %7.2f\n", TW);
		printf ("  Base Diameter = %7.2f\n", TR);
	} else if (IT == eHFrame) {
		printf ("H-Frame:\n");
		printf ("  Height              = %7.2f\n", TW);
		printf ("  Downlead Diameter   = %7.2f\n", TR);
		printf ("  Downlead Separation = %7.2f\n", WW);
	} else if (IT == eCylindrical) {
		printf ("Cylinder:\n");
		printf ("  Height      = %7.2f\n", TW);
		printf ("  Mid Width   = %7.2f\n", WW);
	} else { /* eWaist */
		printf ("Waist:\n");
		printf ("  Height         = %7.2f\n", TW);
		printf ("  Base Diameter  = %7.2f\n", TR);
		printf ("  Top Diameter   = %7.2f\n", TOPR);
		printf ("  Waist-to-Top   = %7.2f\n", TPH);
	}
}

void EchoInput (const char *label) {
	int i;

	printf ("%s\n\n", title);

	if (label && label[0]) {
		printf ("%s\n", label);
	}

	if (THD > 0.0) {
		printf ("Thunderday Level = %7.2f\n", THD);
	} else {
		printf ("Ground Flash Density = %7.2f\n", GFD);
	}

	EchoConductors (NULL);

	printf ("Footing Resistance Distribution:\n");
	for (i = 0; i < NFR; i++) {
		printf ("%7.2f%% - %12.2f ohms\n", 100.0 * PRF[i], RF[i]);
	}

	printf ("\n");
}

int main (int argc, char *argv[]) {
	FILE *fp = NULL;
	char *tok;
	char c;
	char fname [FILENAME_MAX + 1];
	int i;
	int bBriefOutput = 0;

	if (argc < 2) {
		printf ("Usage: flash [-b] case1.dat\n");
		printf ("\tinput data in case1.dat\n");
		printf ("\t-b for brief output\n");
		printf ("\tredirect output like \" > case1.out\" to create an output file.\n");
		exit (EXIT_FAILURE);
	}

	Reset ();
	fname[0] = '\0';

	for (i = 1; i < argc; i++) {
		tok = argv[i];
		if ((tok[0] == '-') && (tok[1] != '\0')) {
			c = tolower (tok[1]);
			if (c == 'b') bBriefOutput = 1;
		} else {
			strncpy (fname, tok, FILENAME_MAX);
		}
	}

	if (!fname[0]) {
		printf ("Error: no input file specified.\n");
		exit (EXIT_FAILURE);
	}

	fp = fopen (fname, "rb");

	if (!fp) {
		printf ("Error: could not open \"%s\"\n", fname);
		exit (EXIT_FAILURE);
	}

	if (!ReadFromFile (fp)) {
		fclose (fp);
		exit (EXIT_FAILURE);
	}

	fclose (fp);

	if (!bBriefOutput) {
		EchoInput ("Input Data:");
	}

	if (MM == eEnglish) {
		EnglishToMetric ();
		if (!bBriefOutput) {
			EchoConductors ("Converted to Metric Units:");
		}
	}

	if (!DoCalculations (bBriefOutput ? 0 : 1)) {
		exit (EXIT_FAILURE);
	}

	if (bBriefOutput) {
		WriteBriefOutput ();
	} else {
		WriteSummaryOutput ();
	}

	return 0;
}